@@ -1,11 +1,24 @@
-from typing import Union
+# Licensed under the Apache License, Version 2.0 (the "License");
+# you may not use this file except in compliance with the License.
+# You may obtain a copy of the License at
+#
+#     http://www.apache.org/licenses/LICENSE-2.0
+#
+# Unless required by applicable law or agreed to in writing, software
+# distributed under the License is distributed on an "AS IS" BASIS,
+# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+# See the License for the specific language governing permissions and
+# limitations under the License.
+# ==============================================================================
+"""Contains functions related to the Activation Functions."""
 
+from typing import Union
 from . import ActivationFunction, Sigmoid, ReLU, Softmax, Linear, Tanh
 
 
 def get_activation_function(activation_function_name: Union[str, None]) -> Union[ActivationFunction, None]:
-    """
-    Returns an activation function object based on the name of the activation function.
+    """Returns an activation function object or None if there is no activation function (the same as using 'Linear')
+
     Args:
         activation_function_name: The name of the activation function. The possible names are:
             - "relu": ReLU
@@ -14,8 +27,6 @@ def get_activation_function(activation_function_name: Union[str, None]) -> Union
             - "tanh": Tanh
             - "linear": Linear
             - None: No activation function (i.e. the input is passed through, the same as Linear)
-    Returns:
-        An activation function object or None if there is no activation function (the same as using 'Linear').
     """
     if activation_function_name is None:
         return None
