@@ -14,10 +14,9 @@
 
 import numpy as np
 from typing import Tuple, Union
-
 from .layer import Layer
 from dlfs.activation_functions import ActivationFunction
-from dlfs.convolutions import Convolutioner, get_convolution
+from dlfs.convolutioners import Convolutioner, get_convolution
 
 
 class Conv2D(Layer):
@@ -27,9 +26,19 @@ class Conv2D(Layer):
     the spatial location. It also concerns some other parameters such as stride, kernel size, padding,
     etc; which are essential characteristics in order to carry out a convolution.
 
+    Technical details (assuming data_format='channels_last'):
+        - The input is a 4D tensor (batch, height, width, number of input channels).
+        - The output is a 4D tensor (batch, height, width, n_filters). (see _get_output_shape for more details)
+        - Each kernel is a 3D tensor (kernel_height, kernel_width, number of input channels).
+        - The weights are a 4D tensor (kernel_height, kernel_width, number of input channels, n_filters).
+        - The bias is a 1D tensor (n_filters,).
+        - The layer performs a '3D' convolution on the input tensor for each kernel.
+
+
+
     Args:
         kernel_size (tuple): tuple of 2 integers, specifying the height and width of the 2D convolution window.
-        n_filters (int): Number of filters
+        n_filters (int): Number of filters/kernels.
         stride (tuple or int): specifying the strides of the convolution along the height and width.
             Can be a single integer to specify the same value for all spatial dimensions
         padding (bool, tuple or int): If True, add padding to the input so that the output has the same shape as the
@@ -38,7 +47,7 @@ class Conv2D(Layer):
             is added to the input on both sides.
         activation (ActivationFunction): Activation function
         use_bias (bool): Whether to use bias
-        convolution_type (str): convolution mode. Can be 'winograd' or 'naive'
+        convolution_type (str): convolution mode. Can be 'winograd', 'direct' or 'patch'.
         name (str): Name of the layer
         input_shape (tuple): shape of the input
         weights_init (str): Initialization method for the weights
@@ -55,7 +64,11 @@ class Conv2D(Layer):
         use_bias (bool): Whether to use bias.
         convolution_type (str): convolution mode. Recommended to be 'winograd'.
         name (str): Name of the layer
+        blocksize (Tuple[int, int]): the size of the block, only used with Winograd.
+        data_format (str): the data format of the input. Can be 'channels_last' or 'channels_first'.
 
+    Raises:
+        ValueError: If using padding and stride != 1.
     """
 
     def __init__(self,
@@ -65,26 +78,28 @@ class Conv2D(Layer):
                  padding: Union[bool, tuple, int] = False,
                  activation: str = None,
                  use_bias: bool = True,
-                 convolution_type: str = 'winograd',
+                 convolution_type: str = "simple",
                  name: str = "Conv2D",
                  input_shape: tuple = None,
-                 weights_init: str = "xavier",
-                 bias_init: str = "zeros"):
-
-        if n_filters <= 0:
-            raise ValueError("The number of filters should be greater than 0")
-        if len(kernel_size) != 2:
-            raise ValueError("The kernel size should be a tuple of two integers")
-        if kernel_size[0] <= 0 or kernel_size[1] <= 0:
-            raise ValueError("The kernel size should be greater than 0")
-        if stride <= 0:
-            raise ValueError("The stride should be greater than 0")
+                 weights_init: str = "glorot_uniform",
+                 bias_init: str = "zeros",
+                 blocksize: Tuple[int, int] = None,
+                 data_format: str = "channels_last"):
 
         if isinstance(kernel_size, int):
             kernel_size = (kernel_size, kernel_size)
         if isinstance(stride, int):
             stride = (stride, stride)
 
+        if n_filters < 0:
+            raise ValueError("The number of filters must be greater than 0")
+
+        if len(kernel_size) != 2:
+            raise ValueError("The kernel size should be a tuple of two integers")
+
+        if kernel_size[0] <= 0 or kernel_size[1] <= 0:
+            raise ValueError("The kernel size should be greater than 0")
+
         input_shape = None if input_shape is None else (None, *input_shape)
         output_shape = None
         if isinstance(padding, bool):
@@ -108,12 +123,15 @@ class Conv2D(Layer):
         self.stride = stride
         self.padding = padding
         self.convolution_type = convolution_type
+        self.blocksize = blocksize
         self.use_bias = use_bias
         self.weights_init = weights_init
         self.bias_init = bias_init
-        self.forward_conv = get_convolution(convolution_type, kernel_size, stride, padding)
+        self.forward_conv = None
         self.backward_conv = None
         self.update_conv = None
+        self.data_format = data_format
+        self._batch_count = 0  # needed to preprocess the data in the winograd algorithm
 
     def initialize(self, input_shape: tuple, weights: np.ndarray = None, bias: np.ndarray = None):
         """Initializes the layer.
@@ -139,19 +157,31 @@ class Conv2D(Layer):
                              "(n_samples, height, width, n_channels)")
 
         self.input_shape = input_shape
+        self.forward_conv = get_convolution(self.convolution_type,
+                                            input_shape[1:],
+                                            self.kernel_size,
+                                            self.padding,
+                                            self.stride,
+                                            self.blocksize)
+
         self.output_shape = self._get_output_shape()
 
-        weights_shape = (input_shape[3], self.kernel_size[0], self.kernel_size[1], self.n_filters)
+        if self.data_format == "channels_first":
+            weights_shape = (self.n_filters, input_shape[3], *self.kernel_size)
+        else:
+            weights_shape = (*self.kernel_size, input_shape[3], self.n_filters)
+
         # initialize weights
         if weights is not None:
             if weights.shape != weights_shape:
                 raise ValueError(f"The shape of the weights should be "
-                                 "(n_channels_prev_layer, kernel_height, kernel_width, n_channels_current_layer). "
+                                 "(n_filters, kernel_height, kernel_width, n_channels_previous_layer). "
                                  f"Got {weights.shape}, expected {weights_shape}")
             self.weights = weights
         elif self.weights_init == "xavier":
             self.weights = np.random.normal(loc=0,
-                                            scale=np.sqrt(1 / (input_shape[3] * self.kernel_size[0] * self.kernel_size[1])),
+                                            scale=np.sqrt(
+                                                1 / (input_shape[3] * self.kernel_size[0] * self.kernel_size[1])),
                                             size=weights_shape)
         elif self.weights_init == "zeros":
             self.weights = np.zeros(weights_shape)
@@ -193,15 +223,34 @@ class Conv2D(Layer):
 
         # INITIALIZE BACKWARD CONVOLUTION
         # compute the padding needed for the backward convolution
-        padding_backward = ((self.kernel_size[0] - 1) // 2, (self.kernel_size[1] - 1) // 2)
+        padding_backward = ((self.kernel_size[0] - 1), (self.kernel_size[1] - 1))
 
         # initialize the backward convolution. This convolution is used to compute the gradient of the loss
-        # with respect to input.
+        # with respect to the input.
         self.backward_conv = get_convolution(self.convolution_type,
                                              image_size=self.output_shape,  # delta.shape
                                              kernel_size=self.kernel_size,
                                              stride=self.stride,
-                                             padding=padding_backward)
+                                             padding=padding_backward,
+                                             data_format=self.data_format)
+
+        # INITIALIZE UPDATE CONVOLUTION
+        # This convolution is used to compute the gradient of the loss with respect to the weights.
+
+        # The image size is the size of the output of the convolution but after applying the padding.
+        if self.data_format == "channels_first":
+            image_height, image_width = (self.output_shape[1], self.output_shape[2])
+        else:
+            image_height, image_width = (self.output_shape[2], self.output_shape[3])
+
+        # compute the image size after applying the padding
+        image_size = (image_height + 2 * padding_backward[0], image_width + 2 * padding_backward[1])
+
+        self.update_conv = get_convolution(self.convolution_type,
+                                           image_size=image_size,
+                                           kernel_size=self.kernel_size,
+                                           stride=self.stride,
+                                           data_format=self.data_format)
 
         self.initialized = True
 
@@ -218,17 +267,18 @@ class Conv2D(Layer):
         """
         if weights is not None:
             # check if the weights shape is correct
-            weight_shape = (self.input_shape[3], self.kernel_size[0], self.kernel_size[1], self.n_filters)
-            if weights.shape != weight_shape:
-                raise ValueError(f"The shape of the weights should be "
-                                 "(n_channels_prev_layer, kernel_height, kernel_width, n_channels_current_layer). "
-                                 f"Got {weights.shape}, expected {weight_shape}")
+            if self.data_format == "channels_first":
+                weights_shape = (self.n_filters, self.input_shape[1], self.kernel_size[0], self.kernel_size[1])
+            else:
+                weights_shape = (self.n_filters, self.kernel_size[0], self.kernel_size[1], self.input_shape[-1])
+            if weights.shape != weights_shape:
+                raise ValueError(f"The shape of the weights should be {weights_shape}, got {weights.shape}.")
             self.weights = weights
 
         if bias is not None:
 
             if not self.use_bias:
-                raise ValueError("The layer does not use bias")
+                raise ValueError("The layer does not use bias.")
 
             # check if the bias shape is correct
             bias_shape = (self.n_filters,)
@@ -239,6 +289,45 @@ class Conv2D(Layer):
 
             self.bias = bias
 
+    def _convolve(self,
+                  x: np.ndarray,
+                  kernel: np.ndarray,
+                  convolutioner: Convolutioner,
+                  data_format: str = "channels_last") -> np.ndarray:
+        """Convolve a batch of multichannel images given a 'Convolutioner'
+
+        Args:
+            x (np.ndarray): batch of images.
+            kernel (np.ndarray): kernel to convolve with.
+            convolutioner (Convolutioner): convolutioner used to convolve the images.
+
+        Returns:
+            np.ndarray: batch of convolved images.
+        """
+
+        if data_format == "channels_last":
+            outputs = np.array([convolutioner.convolve(x,
+                                                       kernel[..., i],
+                                                       batch_count=self._batch_count,
+                                                       data_format=data_format) + self.bias[i]
+                                for i in range(kernel.shape[-1])])
+
+            # the shape of self.outputs.shape is (n_filters, batch_size, height, width). So we need to move the axis
+            # to (batch_size, height, width, n_filters)
+            outputs = np.moveaxis(outputs, 0, -1)
+        else:  # data_format == "channels_first"
+            outputs = np.array([convolutioner.convolve(x,
+                                                       kernel[i],
+                                                       batch_count=self._batch_count,
+                                                       data_format=data_format) + self.bias[i]
+                                for i in range(kernel.shape[0])])
+
+            # the shape of self.outputs.shape is (n_filters, batch_size, height, width). So we need to move the axis
+            # to (batch_size, n_filters, height, width)
+            outputs = np.moveaxis(outputs, 0, 1)
+
+        return outputs
+
     def forward(self, x: np.ndarray, training: bool = False) -> np.ndarray:
         """Forward pass
 
@@ -250,7 +339,14 @@ class Conv2D(Layer):
             output data
         """
         self.inputs = x
-        self.outputs = self.forward_conv.convolve(x, self.weights) + self.bias
+        self.outputs = self._convolve(x, self.weights, self.forward_conv, data_format=self.data_format)
+
+        # update the batch count
+        self._batch_count += 1
+        # Take the modulo 2 to reduce memory usage. We only need to pass a batch count different from the previous one
+        # to the update convolution, in order to indicate that the batch has changed. That's why we can do this.
+        self._batch_count %= 2
+
         return self.outputs if self.activation is None else self.activation(self.outputs)
 
     def get_d_inputs(self, delta: np.ndarray) -> np.ndarray:
@@ -259,17 +355,42 @@ class Conv2D(Layer):
         Args:
             delta: derivative of the cost function with respect to the output of the layer.
         """
-        flipped_weights = np.rot90(self.weights, 2, axes=(1, 2))  # 180° rotation of the filters
-        return self.backward_conv.convolve(delta, flipped_weights)
+        # 180° rotation of the filters
+        rotated_weights = np.rot90(self.weights, 2, axes=(1, 2)) if self.data_format == "channels_last" else \
+            np.rot90(self.weights, 2, axes=(2, 3))
+        a = 0
+        a += 1
+
+        # The current shape of the rotated weights is (height, width, n_input_channels, n_output_channels).
+        # We need to exchange the number of input and output channels because we are in the backward pass.
+        if self.data_format == "channels_last":
+            rotated_weights = np.moveaxis(rotated_weights, 2, -1)
+        else:
+            rotated_weights = np.moveaxis(rotated_weights, 2, 0)
+            rotated_weights = np.moveaxis(rotated_weights, -1, 0)
+
+        d_inputs = self._convolve(delta, rotated_weights, self.backward_conv, data_format=self.data_format)
+        return d_inputs
 
     def count_params(self) -> int:
         return self.weights.size + self.bias.size
 
-    def _get_output_shape(self) -> tuple:
-        return (self.input_shape[0],
-                self.input_shape[1] - self.kernel_size[0] + 1,
-                self.input_shape[2] - self.kernel_size[1] + 1,
-                self.n_filters)
+    def _get_output_shape(self) -> Tuple[int, int, int, int]:
+        """Returns the shape of the output of the layer, taking into account the already defined attributes.
+
+        If the convolution mode is not 'winograd', the output shape could be computed as usual:
+        batch_size = self.input_shape[0]
+        output_height = (self.input_shape[1] - self.kernel_size[0] + 2 * self.padding[0]) // self.stride[0] + 1
+        output_width = (self.input_shape[2] - self.kernel_size[1] + 2 * self.padding[1]) // self.stride[1] + 1
+
+        return batch_size, output_height, output_width, self.n_filters
+
+        However, if the convolution mode is 'winograd', the output shape is computed differently (see the
+        `get_output_shape` method of the `WinogradConvolutioner` class for more details). For
+         this reason we need to call the `get_output_shape` method of the `Convolution` class."""
+        batch_size = self.input_shape[0]
+        output_height, output_width = self.forward_conv.get_output_shape()
+        return batch_size, output_height, output_width, self.n_filters
 
     def update(self, optimizer, delta: np.ndarray):
         """Updates the weights and biases of this layer.
@@ -284,10 +405,32 @@ class Conv2D(Layer):
         if not self.initialized:
             raise ValueError("The layer is not initialized")
 
-        dw = self.update_conv.convolve(self.inputs, delta.transpose((0, 3, 2, 1)))
-        db = np.sum(delta, axis=(1, 2))
+        # Compute the gradient of the loss with respect to the weights and biases
+
+        if self.data_format == "channels_last":
+            # The self.inputs.shape is (batch_size, inputs_height, inputs_width, n_channels)
+            # The delta.shape is (batch_size, inputs_height, inputs_width, n_filters)
+            # The self.weights.shape is (kernel_height, kernel_width, n_channels, n_filters)
+
+            # The gradient of the loss with respect to the weights is computed as:
+            # d_w = convolve(inputs, delta)
+
+            # We can not perform this operation directly because the convolution function expects the inputs to
+            # have the shape (batch_size, inputs_height, inputs_width, n_channels) and the kernel (deltas) to have
+            # the shape (kernel_height, kernel_width, n_channels). Thus, we have to use a list comprehension to
+            # perform 'n_filters*batch_size' convolutioners and then sum the results and move the axes to the correct
+            # positions.
+
+            dw = np.zeros(self.weights.shape)
+            for i, image in enumerate(self.inputs):
+                for j in range(image.shape[2]):  # image.shape[2] = number of channels
+                    for k in range(self.n_filters):
+                        dw[:, :, j, k] += self.update_conv(image[:, :, j], delta[i, :, :, k], using_batches=False)
+
+            dw = dw / self.inputs.shape[0]
+            db = np.sum(delta, axis=(0, 1, 2))
 
         optimizer.update(self, (dw, db))
 
     def summary(self):
-        print(f"Layer: {self.name}, Output shape: {self.output_shape}")
+        return f"Conv2D: {self.name}, Output shape: {self.output_shape}"
